package com.sample;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.drools.KnowledgeBase;
import org.drools.KnowledgeBaseFactory;
import org.drools.builder.KnowledgeBuilder;
import org.drools.builder.KnowledgeBuilderError;
import org.drools.builder.KnowledgeBuilderErrors;
import org.drools.builder.KnowledgeBuilderFactory;
import org.drools.builder.ResourceType;
import org.drools.io.ResourceFactory;
import org.drools.logger.KnowledgeRuntimeLogger;
import org.drools.logger.KnowledgeRuntimeLoggerFactory;
import org.drools.runtime.StatefulKnowledgeSession;

/**
 * This is a sample class to launch a rule.
 */
public class DroolsTest {

	public static final void main(String[] args) {
		try {
			// load up the knowledge base
			KnowledgeBase kbase = readKnowledgeBase();
			StatefulKnowledgeSession ksession = kbase.newStatefulKnowledgeSession();
			KnowledgeRuntimeLogger logger = KnowledgeRuntimeLoggerFactory.newFileLogger(ksession, "test");
			// go !
			
			Data data = new Data(0);
			data.add(new Criteria(data, "c0", "foo"));
			data.add(new Criteria(data, "c1", "bar"));
			data.add(new Criteria(data, "c2", "bas"));
			ksession.insert(data);

			{//Exact match
				Query query = new Query(0);
				query.add(new Criteria(query, "c0", "foo"));
				query.add(new Criteria(query, "c1", "bar"));
				query.add(new Criteria(query, "c2", "bas"));
				ksession.insert(query);
				for(Criteria c : query) {
					ksession.insert(c);
				}
			}
			{//Match
				Query query = new Query(1);
				query.add(new Criteria(query, "c0", "foo"));
				query.add(new Criteria(query, "c1", "bar"));
				ksession.insert(query);
				for(Criteria c : query) {
					ksession.insert(c);
				}
			}
			{//Mismatch
				Query query = new Query(2);
				query.add(new Criteria(query, "c0", "foo"));
				query.add(new Criteria(query, "c1", "bar"));
				query.add(new Criteria(query, "c3", "bas"));
				ksession.insert(query);
				for(Criteria c : query) {
					ksession.insert(c);
				}
			}
			{//Mismatch
				Query query = new Query(3);
				query.add(new Criteria(query, "c0", "foo"));
				query.add(new Criteria(query, "c1", "bar"));
				query.add(new Criteria(query, "c2", "bas"));
				query.add(new Criteria(query, "c3", "bing"));
				ksession.insert(query);
				for(Criteria c : query) {
					ksession.insert(c);
				}
			}
			
			ksession.fireAllRules();
			logger.close();
		} catch (Throwable t) {
			t.printStackTrace();
		}
	}

	private static KnowledgeBase readKnowledgeBase() throws Exception {
		KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder();
		kbuilder.add(ResourceFactory.newClassPathResource("Sample.drl"), ResourceType.DRL);
		KnowledgeBuilderErrors errors = kbuilder.getErrors();
		if (errors.size() > 0) {
			for (KnowledgeBuilderError error: errors) {
				System.err.println(error);
			}
			throw new IllegalArgumentException("Could not parse knowledge.");
		}
		KnowledgeBase kbase = KnowledgeBaseFactory.newKnowledgeBase();
		kbase.addKnowledgePackages(kbuilder.getKnowledgePackages());
		return kbase;
	}

	public static class Criteria {
		public final Object parent;
		public final String name;
		public final Object value;
		public Criteria(Object parent, String name, Object value) {
			super();
			this.parent = parent;
			this.name = name;
			this.value = value;
		}
		public Object getParent() {
			return parent;
		}
		public String getName() {
			return name;
		}
		public Object getValue() {
			return value;
		}
		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((name == null) ? 0 : name.hashCode());
			result = prime * result + ((value == null) ? 0 : value.hashCode());
			return result;
		}
		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			Criteria other = (Criteria) obj;
			if (name == null) {
				if (other.name != null)
					return false;
			} else if (!name.equals(other.name))
				return false;
			if (value == null) {
				if (other.value != null)
					return false;
			} else if (!value.equals(other.value))
				return false;
			return true;
		}
		
		public String toString() {
			return name + "->" + value;
		}
	}
	
	public static class Data implements Collection<Criteria>{
		
		private final int id;
		private final Map<String, Criteria> values = new HashMap<String, Criteria>();
		
		public Data(int id) {
			this.id = id;
		}

		public int getId() {
			return id;
		}
		
		public Map<String, Criteria> getValues() {
			return values;
		}

		public boolean add(Criteria e) {
			Criteria old = values.put(e.getName(), e);
			if(old == null) {
				return true;
			} else {
				return (old.value != null && !old.value.equals(e.value)) || (old.value == null && e.value != null);
			}
		}

		public boolean addAll(Collection<? extends Criteria> c) {
			boolean changed = false;
			for(Criteria crit : c) changed |= add(crit);
			return changed;
		}

		public void clear() {
			values.clear();
		}

		public boolean contains(Object o) {
			return values.values().contains(o);
		}

		public boolean containsAll(Collection<?> c) {
			return values.values().containsAll(c);
		}

		public boolean isEmpty() {
			return values.isEmpty();
		}

		public Iterator<Criteria> iterator() {
			return values.values().iterator();
		}

		public boolean remove(Object o) {
			return values.values().remove(o);
		}

		public boolean removeAll(Collection<?> c) {
			return values.values().removeAll(c);
		}

		public boolean retainAll(Collection<?> c) {
			return values.values().retainAll(c);
		}

		public int size() {
			return values.size();
		}

		public Object[] toArray() {
			return values.values().toArray();
		}

		public <T> T[] toArray(T[] a) {
			return values.values().toArray(a);
		}
		
		public String toString() {
			return "Data" + id + " " + values;
		}
	}
	
	public static class Query implements Collection<Criteria>{
		
		private final int id;
		private final Map<String, Criteria> values = new HashMap<String, Criteria>();
		
		public Query(int id) {
			this.id = id;
		}

		public int getId() {
			return id;
		}
		
		public Map<String, Criteria> getValues() {
			return values;
		}

		public boolean add(Criteria e) {
			Criteria old = values.put(e.getName(), e);
			if(old == null) {
				return true;
			} else {
				return (old.value != null && !old.value.equals(e.value)) || (old.value == null && e.value != null);
			}
		}

		public boolean addAll(Collection<? extends Criteria> c) {
			boolean changed = false;
			for(Criteria crit : c) changed |= add(crit);
			return changed;
		}

		public void clear() {
			values.clear();
		}

		public boolean contains(Object o) {
			if(o instanceof Criteria) {
				Criteria c = (Criteria)o;
				return c.parent == this && values.values().contains(o);
			} else {
				return false;
			}
		}

		public boolean containsAll(Collection<?> c) {
			return values.values().containsAll(c);
		}

		public boolean isEmpty() {
			return values.isEmpty();
		}

		public Iterator<Criteria> iterator() {
			return values.values().iterator();
		}

		public boolean remove(Object o) {
			return values.values().remove(o);
		}

		public boolean removeAll(Collection<?> c) {
			return values.values().removeAll(c);
		}

		public boolean retainAll(Collection<?> c) {
			return values.values().retainAll(c);
		}

		public int size() {
			return values.size();
		}

		public Object[] toArray() {
			return values.values().toArray();
		}

		public <T> T[] toArray(T[] a) {
			return values.values().toArray(a);
		}
		
		public String toString() {
			return "Query" + id + " " + values;
		}
	}
}